from pandas import read_csv as pd_read_csv
from argparse import ArgumentParser
from numpy import array_split as np_array_split
from os import makedirs as os_makedirs
from os.path import join as os_path_join, basename as os_path_basename, splitext as os_path_splitext

def split_chrom_tsv(chrom_tsv, output_folder, split_num=0):
  chrom = pd_read_csv(chrom_tsv, sep='\t')
  chrom.columns = chrom.columns.str.replace(' ', '')
  pep_list = chrom['PeptideModifiedSequence'].unique()
  by_column = 'PeptideModifiedSequence'
  os_makedirs(output_folder, exist_ok=True)
  prefix = os_path_join(output_folder,os_path_basename(os_path_splitext(chrom_tsv)[0]))
  if split_num == 0:
    splited_list = np_array_split(chrom[by_column].unique(), len(pep_list))
  else:
    splited_list = np_array_split(chrom[by_column].unique(), split_num)
  for i in range(len(splited_list)):
    data = chrom[chrom[by_column].isin(splited_list[i])]
    output_path = prefix + f'_split_{i}.tsv'
    row_num = len(data)
    print(f'Output to {output_path} containing {row_num} rows.')
    data.to_csv(output_path, sep='\t', index=False)

def main():
  parser = ArgumentParser()
  parser.add_argument("chrom_tsv", type=str, help="The chromatogram tsv file path")
  parser.add_argument("output_folder", type=str, help="The output folder of splitted chromatogram files")
  parser.add_argument("--number", '-n', type=int, default=0, help="The desired slice number of the chromatogram tsv files")
  args = parser.parse_args()
  split_chrom_tsv(args.chrom_tsv, args.output_folder, split_num=args.number)

if __name__ == '__main__':
  main()


from argparse import ArgumentParser
from glob import glob
from numpy import array as np_array, median as np_median, inf as np_inf
from pandas import read_csv as pd_read_csv, concat as pd_concat
from os import stat as os_stat
from os.path import join as os_path_join


def merge_peak_csv(input_folder, output_csv, picked=None, suffix=None):
  if suffix:
    csv_files = glob(os_path_join(input_folder, f'*{suffix}.csv'))
  else:
    csv_files = glob(os_path_join(input_folder, '*.csv'))
  sorted_csv_files = sorted(csv_files)
  print(f'Merge peak boundary csv files: {sorted_csv_files}')
  dataframes = map(pd_read_csv, filter(lambda x: os_stat(x).st_size > 1, sorted_csv_files))
  output_df = pd_concat(dataframes, ignore_index=True)

  print('#########################################################')
  print('Score distributions:')
  total_len = len(output_df)
  if total_len > 0:
    ranges = [-1, 0, 0.01, 1, 2, 4, 6, 8, 10, 12]
    for idx, range_start in enumerate(ranges):
      if idx + 1 < len(ranges):
        range_end = ranges[idx + 1]
      else:
        range_end = np_inf
      in_range_len = len(output_df[(output_df['FinalReward'] >= range_start) & (output_df['FinalReward'] < range_end)])
      print(f'[{range_start}, {range_end}): {in_range_len} ({ round(100*in_range_len/total_len, 2) }%)')
  print('#########################################################')
  print(f'Output merged csv file to {output_csv}')
  output_df.to_csv(output_csv, index=False)

  if picked is not None:
    pre_picked = pd_read_csv(picked)
    if 'FinalReward' in pre_picked:
      print('#########################################################')
      print('Comparing with previous picked peaks:')
      improved_counts = 0
      improved_scores = []
      for idx, row in pre_picked.iterrows():
        fn = row['File Name']
        ps = row['Peptide Modified Sequence']
        r = row['FinalReward']
        matched_row = output_df[(output_df['File Name'] == fn) & (output_df['Peptide Modified Sequence'] == ps)]
        if not matched_row.empty and len(matched_row) == 1:
          new_r = matched_row['FinalReward'].tolist()[0]
        if new_r > r:
          improved_counts += 1
          improved_scores.append(new_r - r)
      if len(improved_scores) > 0:
        improved_scores = np_array(improved_scores)
        improved_mean = improved_scores.mean()
        improved_median = np_median(improved_scores)
        improved_min = improved_scores.min()
        improved_max = improved_scores.max()
        print(f'Improved chromatograms: {improved_counts}')
        print(f'Improved scores => mean: {improved_mean}; median: {improved_median}; min: {improved_min}; max: {improved_max}')
      else:
        print('No improvement')
      print('#########################################################')

  print('Finished.')

def main():
  parser = ArgumentParser()
  parser.add_argument('input_folder', type=str, help='The input folder containing splitted chromatogram files and picked peak csv fils.')
  parser.add_argument('output_csv', type=str, help='The output csv file of the merged peak data.')
  parser.add_argument('--picked', type=str, default=None, help="The previously picked peak boundaries to compare and summarize the differences. (default: None)")
  parser.add_argument('--suffix', type=str, default=None, help="The filename suffix pattern of csv files. (default: None)")
  args = parser.parse_args()
  merge_peak_csv(args.input_folder, args.output_csv, picked=args.picked, suffix=args.suffix)

if __name__ == '__main__':
  main()


from itertools import groupby
from .QualityEncoder import QualityEncoder
from numpy import (sort as np_sort, median as np_median, mean as np_mean, average as np_average, max as np_max, argmax as np_argmax, min as np_min, std as np_std, arange as np_arange, sqrt as np_sqrt, sum as np_sum, power as np_power,
                  nan as np_nan, isnan as np_isnan, nan_to_num as np_nan_to_num, abs as np_abs, zeros as np_zeros, fill_diagonal as np_fill_diagonal, floor as np_floor,
                  diff as np_diff, sign as np_sign, corrcoef as np_corrcoef, argsort as np_argsort, flip as np_flip,
                  hstack as np_hstack, vstack as np_vstack, concatenate as np_concatenate,
                  array as np_array, repeat as np_repeat, tile as np_tile, transpose as np_transpose, 
                  vsplit as np_vsplit, hsplit as np_hsplit, split as np_split,
                  errstate as np_errstate, where as np_where, apply_along_axis as np_apply_along_axis)

class PeakQualityEval():
  def __init__(self, device='auto'):
    self.flatness_factor = 0.05
    self.quality_encoder = QualityEncoder(device=device)

  def top_n_frag_idx_list(self, chrom_data, top_n=5):
    (_, heavy_peak_ints) = np_hsplit(chrom_data[4], 2)
    if heavy_peak_ints.shape[0] > 0:
      heavy_max_peak_ints = np_max(heavy_peak_ints, axis=0)
      transition_idx = np_argsort(heavy_max_peak_ints)[-top_n:]
      return np_flip(transition_idx)
    else:
      return np_arange(0, len(chrom_data[2][0]))[-top_n:]

  def getOverallScore(self, chrom_data, top_n=5):
    type1_quality_list = self.batchGetType1Quality([chrom_data])
    type1_scores =  self.quality_encoder(type1_quality_list[0])[0]
    (_, area2sum_ratio_heavy)  = np_hsplit(chrom_data[6], 2)
    q1_score_idx_list = np_flip(np_argsort(area2sum_ratio_heavy)[-top_n:])
    q1_score = np_median(type1_scores[q1_score_idx_list])
    if len(type1_scores) > top_n:
      (_, heavy_peak_ints) = np_hsplit(chrom_data[4], 2)
      if heavy_peak_ints.shape[0] > 0:
        heavy_max_peak_ints = np_max(heavy_peak_ints, axis=0)
        transition_idx = np_argsort(heavy_max_peak_ints)[-top_n:]
        transition_idx = np_concatenate((transition_idx, transition_idx + len(chrom_data[2][0])))
        type2_quality = self.makeType2Feature(chrom_data[4][:, transition_idx], chrom_data[5], chrom_data[6][transition_idx])
        q2_score = self.quality_encoder(np_array([type2_quality]))[0][0]
      else:
        type2_quality = self.batchGetType2Quality([chrom_data])
        q2_score = self.quality_encoder(np_array(type2_quality))[0][0]
    else:
      type2_quality = self.batchGetType2Quality([chrom_data])
      q2_score = self.quality_encoder(np_array(type2_quality))[0][0]
    return q1_score, q2_score
  
  def batchGetType1Score(self, chrom_data_list):
    type1_quality_list = self.batchGetType1Quality(chrom_data_list)
    return np_hstack([self.quality_encoder(q)[0] for q in type1_quality_list])
  def batchGetType2Score(self, chrom_data_list, top_n=None):
    type2_quality = self.batchGetType2Quality(chrom_data_list, top_n=top_n)
    return self.quality_encoder(np_array(type2_quality))[0]
  def batchGetType3Score(self, chrom_data_list):
    type3_quality_list = self.batchGetType3Quality(chrom_data_list)
    return np_hstack([self.quality_encoder(q)[0] for q in type3_quality_list])

  def batchGetType1Quality(self, chrom_data_list): #zip(peak_intensity_list, peak_time_list, area_sum_ratio_list)
    return [
      np_hstack(
        [np_hstack(np_vsplit(np_array([
          self.calculateTransitionJaggedness(x[4]),
          self.calculateTransitionSymmetry(x[4]),
          self.calculateTransitionModality(x[4]),
          self.calculateTransitionShift(x[4], x[5]),
          self.calculateTransitionFWHM(x[4], x[5]),
          self.calculateTransitionFWHM2base(x[4], x[5]),
          self.calculateTransitionMaxBoundaryIntensityNormalized(x[4])
        ]).T, 2)), 
        np_array([
          self.calculatePairSimilarity(x[4]),
          self.calculatePairShift(x[4], x[5]),
          self.calculatePairFWHMConsistency(x[4], x[5])
        ]).T])
      for x in chrom_data_list
    ]
  def processForType2(self, chrom_data, top_n):
      transition_idx = self.top_n_frag_idx_list(chrom_data, top_n)
      transition_idx = np_concatenate((transition_idx, transition_idx + len(chrom_data[2][0])))
      return self.makeType2Feature(chrom_data[4][:, transition_idx], chrom_data[5], chrom_data[6][transition_idx])
  def batchGetType2Quality(self, chrom_data_list, top_n = None):
    if top_n is None:
        return [self.makeType2Feature(x[4], x[5], x[6]) for x in chrom_data_list]
    else:
        return [self.processForType2(x, top_n) for x in chrom_data_list]
  def makeType2Feature(self, peak_ints, peak_time, area_ratio):
    return np_hstack([
      np_hstack(np_vsplit(np_array([
        self.calculateIsotopeJaggedness(peak_ints),
        self.calculateIsotopeSymmetry(peak_ints),
        self.calculateIsotopeSimilarity(peak_ints),
        self.calculateIsotopeModality(peak_ints),
        self.calculateIsotopeShift(peak_ints, peak_time),
        self.calculateIsotopeFWHM(peak_ints, peak_time),
        self.calculateIsotopeFWHM2base(peak_ints, peak_time)
      ]).T, 2))[0],
      np_array([
        self.calculatePeakGroupRatioCorr(area_ratio),
        self.calculatePeakGroupJaggedness(peak_ints),
        self.calculatePeakGroupSymmetry(peak_ints),
        self.calculatePeakGroupSimilarity(peak_ints),
        self.calculatePeakGroupModality(peak_ints),
        self.calculatePeakGroupShift(peak_ints, peak_time),
        self.calculatePeakGroupFWHM(peak_ints, peak_time),
        self.calculatePeakGroupFWHM2base(peak_ints, peak_time)
      ])
    ])
  def getType3Quality(self, grouped):
    file_number = len(grouped)
    result = np_hstack([
        np_transpose(np_hsplit(self.calculateMeanIsotopeRatioConsistency(grouped), 2), axes=(0, 2, 1)).reshape(2, 1, -1).squeeze(1).T,
        np_transpose(np_hsplit(self.calculateMeanIsotopeFWHMConsistency(grouped), 2), axes=(0, 2, 1)).reshape(2, 1, -1).squeeze(1).T,
        np_transpose(np_repeat(np_hsplit(self.calculateArea2SumRatioCV(grouped), 2), file_number).reshape(2, -1))
    ])
    result = np_hstack([
      result[:, ::2],  # Light
      result[:, 1::2], # Heavy
      np_array([
        np_tile(self.calculateMeanIsotopeRTConsistency(grouped), len(grouped[0][2][0])).T,
        self.calculatePairRatioConsistency(grouped).transpose().reshape(1, -1)[0]
      ]).T
    ])
    return np_vstack([result[i::file_number, :] for i in range(file_number)]) # ordering: 1) filename; 2) transition ions
  def batchGetType3Quality(self, chrom_data_list):
    hasVariedTransitions = False
    mol2transitions = {}
    sorted_chrom_data_list = sorted(chrom_data_list, key=lambda x: x[0])
    for mol, chrom_list in groupby(sorted_chrom_data_list, key=lambda x: x[0]):
      trans_set = None
      for chrom in chrom_list:
        if trans_set is None:
          trans_set = set(chrom[2][0])
          mol2transitions[mol] = trans_set
          continue
        chrom_trans = set(chrom[2][0])
        mol2transitions[mol] = mol2transitions[mol].union(chrom_trans)
        if len(trans_set) != len(chrom_trans) or trans_set != chrom_trans:
          hasVariedTransitions = True
          break
      if hasVariedTransitions:
        break
    if not hasVariedTransitions:
      return [self.getType3Quality(tuple(x[1])) for x in groupby(sorted_chrom_data_list, key=lambda x: x[0])] # chrom_data_list should be sorted by molecules
    else:
      print('Varied transition numbers are currently unsupported')

# TargetedMSQC metric calculations
# Intensity(1)
  def calculateTransitionMaxBoundaryIntensityNormalized(self, peak_intensity_matrix): #Level 1
    # Set the upperbound value to 5  (the worst case value is 5)
    if peak_intensity_matrix.shape[0] > 0:
      transition_max_boundary_intensity = peak_intensity_matrix[[0, -1], :].max(axis=0)  
      transition_max_intensity = np_max(peak_intensity_matrix, axis=0)
      with np_errstate(divide='ignore', invalid='ignore'):
        TransitionMaxBoundaryIntensityNormalized = transition_max_boundary_intensity / transition_max_intensity
    else:
      TransitionMaxBoundaryIntensityNormalized = np_array([np_nan] * peak_intensity_matrix.shape[1])
    np_nan_to_num(TransitionMaxBoundaryIntensityNormalized, copy=False, nan=5.0)
    TransitionMaxBoundaryIntensityNormalized = np_where(TransitionMaxBoundaryIntensityNormalized < 5, TransitionMaxBoundaryIntensityNormalized, 5)
    return TransitionMaxBoundaryIntensityNormalized

# Shift(4)
  def calculatePairShift(self, peak_intensity_matrix, peak_time): #4 Level 1
    if len(peak_intensity_matrix) == 0 or len(peak_time) <= 1:
      return [1] * int(peak_intensity_matrix.shape[1]/2)
    paired_peak_intensity = np_hsplit(peak_intensity_matrix, 2)
    peak_height_indices = np_argmax(paired_peak_intensity, axis=1)
    PairShift = np_abs(peak_time[peak_height_indices][0] - peak_time[peak_height_indices][1]) / (peak_time[-1]-peak_time[0])
    return PairShift
  def calculateTransitionShift(self, peak_intensity_matrix, peak_time): #8 Level 1
    transition_ion_number = peak_intensity_matrix.shape[1]
    if len(peak_intensity_matrix) < 2:
      return np_array([1] * transition_ion_number)
    peak_max_time = peak_time[peak_intensity_matrix.argmax(axis=0)]
    peak_max_median = np_median(peak_max_time)
    return np_abs(peak_max_time - peak_max_median) / (peak_time[-1]-peak_time[0])
  def calculateIsotopeShift(self, peak_intensity_matrix, peak_time): #2 Level 2
    transition_shift = self.calculateTransitionShift(peak_intensity_matrix, peak_time)
    return np_mean(np_hsplit(transition_shift, 2), axis=1)
  def calculatePeakGroupShift(self,peak_intensity_matrix, peak_time): #1 #Level 2
    transition_shift = self.calculateTransitionShift(peak_intensity_matrix, peak_time)
    return np_mean(transition_shift)

# Jaggedness(3)
  def calculateTransitionJaggedness(self, peak_intensity_matrix=np_array([[]])):
    if peak_intensity_matrix.shape[0] <= 2:
      return np_array([1]*peak_intensity_matrix.shape[1])
    peak_diff_arr = np_diff(peak_intensity_matrix, axis=0)
    peak_diff_arr[np_abs(peak_diff_arr) < self.flatness_factor * np_max(peak_intensity_matrix, axis=0)] = 0
    with np_errstate(divide='ignore', invalid='ignore'):
      jaggedness = ((np_abs(np_diff(np_sign(peak_diff_arr), axis=0)) > 1).sum(axis=0) - 1) / (len(peak_diff_arr)-1)
    np_nan_to_num(jaggedness, copy=False, nan=1)
    jaggedness = np_where(jaggedness > 0, jaggedness, 0)
    return jaggedness
  def calculateIsotopeJaggedness(self, peak_intensity_matrix):
    jaggedness = self.calculateTransitionJaggedness(peak_intensity_matrix)
    isotope_jaggedness = np_mean(np_split(jaggedness, 2), axis=1)
    return isotope_jaggedness
  def calculatePeakGroupJaggedness(self, peak_intensity_matrix): #1 #Level 2
    jaggedness = self.calculateTransitionJaggedness(peak_intensity_matrix)
    return np_mean(jaggedness)
  # Similarity(3)
  def calculatePairSimilarityMatrix(self, peak_intensity_matrix): #matrix
    transition_ion_number = peak_intensity_matrix.shape[1]
    if peak_intensity_matrix.shape[0] < 2:
      PairSimilarityMx = np_zeros((transition_ion_number, transition_ion_number))
      np_fill_diagonal(PairSimilarityMx, 1)
    else:
      with np_errstate(divide='ignore', invalid='ignore'):
        PairSimilarityMx = np_corrcoef(peak_intensity_matrix,  rowvar=False)
    np_nan_to_num(PairSimilarityMx, copy=False, nan=0)
    return PairSimilarityMx
  def calculatePairSimilarity(self, peak_intensity_matrix): #4 Level 1
    similarity_mx = self.calculatePairSimilarityMatrix(peak_intensity_matrix)
    frag_ion_number = int(peak_intensity_matrix.shape[1]/2)
    return similarity_mx[frag_ion_number:, :frag_ion_number].diagonal()
  def calculateIsotopeSimilarity(self, peak_intensity_mx): #2 #Level 2
    similarity_mx = self.calculatePairSimilarityMatrix(peak_intensity_mx)
    frag_num = int(peak_intensity_mx.shape[1]/2)
    light_mean = similarity_mx[:frag_num, :frag_num].mean()
    heavy_mean = similarity_mx[frag_num:, frag_num:].mean()
    return np_array([light_mean, heavy_mean])
  def calculatePeakGroupSimilarity(self, peak_intensity_mx): #1 #Level 2
    similarity_mx = self.calculatePairSimilarityMatrix(peak_intensity_mx)
    return np_mean(similarity_mx)
  
# Symmetry(3)
  def calculateTransitionSymmetry(self, peak_intensity_matrix): #8 #Level 1
    timepoint_number = np_floor(peak_intensity_matrix.shape[0]/2).astype(int)
    if timepoint_number <= 2:
      return np_array([0] * peak_intensity_matrix.shape[1])
    left = peak_intensity_matrix[:(timepoint_number)]
    right = peak_intensity_matrix[:(timepoint_number):-1]
    min_len = np_min((left.shape[0], right.shape[0])) # in case that the length is an odd number
    with np_errstate(divide='ignore', invalid='ignore'):
      corr_mx = np_corrcoef(left[:min_len,:], right[:min_len,:], rowvar=False)
    np_nan_to_num(corr_mx, copy=False, nan=0)
    return np_vsplit(corr_mx, 2)[1].diagonal()
  def calculateIsotopeSymmetry(self, peak_intensity_matrix): #2 #Level 2
    symmetry_arr = self.calculateTransitionSymmetry(peak_intensity_matrix)
    return np_mean(np_hsplit(symmetry_arr, 2), axis=1)
  def calculatePeakGroupSymmetry(self, peak_intensity_matrix): #1 #Level 2
    symmetry_arr = self.calculateTransitionSymmetry(peak_intensity_matrix)
    return np_mean(symmetry_arr)

# FWHM(8)
  def calc_fwhm(self, sig, time):
    peakmax = np_max(sig)
    higher_than_halfpeak = np_where((sig - peakmax/2) > 0)[0]
    try:
      left_index = np_array([higher_than_halfpeak[0] - 1, higher_than_halfpeak[0]])
    except:
      left_index = np_array([np_nan, np_nan])
    try:
      right_index = np_array([higher_than_halfpeak[-1], higher_than_halfpeak[-1] + 1])
    except:
      right_index = np_array([np_nan, np_nan])
    if (left_index[0] == -1) or (np_isnan(left_index[0])):
      t_left = time[0]
    else:
      t_left = (time[left_index[1]] - time[left_index[0]])/(sig[left_index[1]] - sig[left_index[0]])*(peakmax/2 - sig[left_index[0]]) + time[left_index[0]]
    if (right_index[1] > (len(time)-1)) or (np_isnan(right_index[1])):
      t_right = time[-1]
    else:
      t_right = (time[right_index[1]] - time[right_index[0]])/(sig[right_index[1]] - sig[right_index[0]])*(peakmax/2 - sig[right_index[0]]) + time[right_index[0]]
    fwhm = t_right - t_left
    return fwhm
  def calculateTransitionFWHM(self, peak_intensity_matrix, peak_time): #8
    if len(peak_time) < 1:
        return np_array([5] * peak_intensity_matrix.shape[1])
    TransitionFWHM = np_apply_along_axis(self.calc_fwhm, 0, peak_intensity_matrix, peak_time)
    TransitionFWHM = np_where(TransitionFWHM < 5, TransitionFWHM, 5) # Set the upperbound value to 5
    return TransitionFWHM
  def calculateTransitionFWHM2base(self, peak_intensity_matrix, peak_time): #8 #Level 1
    TransitionFWHM = self.calculateTransitionFWHM(peak_intensity_matrix, peak_time)
    TransitionFWHM2Base = TransitionFWHM/(peak_time[-1] - peak_time[0]) if len(peak_time) >= 2 else TransitionFWHM
    return np_where(TransitionFWHM2Base < 5, TransitionFWHM2Base, 5) # Set the upperbound value to 5
  def calculateIsotopeFWHM(self, peak_intensity_matrix, peak_time): #2 #Level 2
    TransitionFWHM = self.calculateTransitionFWHM( peak_intensity_matrix, peak_time)
    IsotopeFWHM = np_mean(np_hsplit(TransitionFWHM, 2), axis=1)
    return np_where(IsotopeFWHM < 5, IsotopeFWHM, 5) # Set the upperbound value to 5
  def calculateIsotopeFWHM2base(self, peak_intensity_matrix, peak_time): #2 Level 2
    TransitionFWHM2Base = self.calculateTransitionFWHM2base(peak_intensity_matrix, peak_time)
    IsotopeFWHM2Base = np_mean(np_hsplit(TransitionFWHM2Base, 2), axis=1)
    return np_where(IsotopeFWHM2Base < 5, IsotopeFWHM2Base, 5) # Set the upperbound value to 5
  def calculatePeakGroupFWHM(self, peak_intensity_matrix, peak_time): #1 #Level 2
    TransitionFWHM = self.calculateTransitionFWHM(peak_intensity_matrix, peak_time)
    PeakGroupFWHM = np_mean(TransitionFWHM)
    return PeakGroupFWHM if PeakGroupFWHM < 5 else 5
  def calculatePeakGroupFWHM2base(self, peak_intensity_matrix, peak_time): #1 #Level 2
    TransitionFWHM2Base = self.calculateTransitionFWHM2base(peak_intensity_matrix, peak_time)
    PeakGroupFWHM2base = np_mean(TransitionFWHM2Base)
    return PeakGroupFWHM2base if PeakGroupFWHM2base < 5 else 5
  def calculatePairFWHMConsistency(self, peak_intensity_matrix, peak_time): #4 Level 1
    TransitionFWHM = self.calculateTransitionFWHM(peak_intensity_matrix, peak_time)
    [lightFWHM, heavyFWHM] = np_hsplit(TransitionFWHM, 2)
    with np_errstate(divide='ignore', invalid='ignore'):
      PairFWHMConsistency = np_abs((lightFWHM - heavyFWHM)/heavyFWHM)
    np_nan_to_num(PairFWHMConsistency, copy=False, nan=5)
    return np_where(PairFWHMConsistency < 5, PairFWHMConsistency, 5) # set the upperbound value to 5
  def calculateMeanIsotopeFWHMConsistency(self, chrom_data_list): #8 # Level 3
    TransitionFWHM_crossAll = np_array([self.calculateTransitionFWHM(x[4], x[5]) for x in chrom_data_list])
    mean = np_mean(TransitionFWHM_crossAll, axis=0)
    with np_errstate(divide='ignore', invalid='ignore'):
      MeanIsotopeFWHMConsistency = np_abs(TransitionFWHM_crossAll - mean)/mean
    np_nan_to_num(MeanIsotopeFWHMConsistency, copy=False, nan=5)
    return np_where(MeanIsotopeFWHMConsistency < 5, MeanIsotopeFWHMConsistency, 5)

# Modality(3)
  def calculateTransitionModality(self, peak_intensity_matrix): #8 #Level 1
    if peak_intensity_matrix.shape[0] < 2:
      return np_array([1]*peak_intensity_matrix.shape[1])
    peak_diff_arr = np_diff(peak_intensity_matrix, axis=0)
    peak_diff_arr[np_abs(peak_diff_arr) < self.flatness_factor * np_max(peak_intensity_matrix, axis=0)] = 0
    def first_fall(series):
      fall_list = np_where(series < 0)[0]
      return fall_list[0] if len(fall_list) > 0 else len(series)
    def last_rise(series):
      rise_list = np_where(series > 0)[0]
      return rise_list[-1] if len(rise_list) > 0 else -1
    first_falls = np_apply_along_axis(first_fall, 0, peak_diff_arr)
    last_rises = np_apply_along_axis(last_rise, 0, peak_diff_arr)
    TransitionModality = []
    for idx, (data_series, f, l) in enumerate(zip(peak_diff_arr.T, first_falls, last_rises)):
      max_dip = 0
      if f < l:
        max_dip = np_max(np_abs(data_series[f:l+1]))
      if len(data_series) == 0:
        modality = 1
      elif np_max(data_series) == 0:
        modality = 1
      else:
        modality = max_dip/np_max(peak_intensity_matrix[:, idx])
      TransitionModality.append(modality)
    return np_array(TransitionModality)
  def calculateIsotopeModality(self, peak_intensity_matrix): #2 #Level 2
    TransitionModality = self.calculateTransitionModality(peak_intensity_matrix)
    return np_mean(np_hsplit(TransitionModality, 2), axis=1)
  def calculatePeakGroupModality(self, peak_intensity_matrix): #1 #Level 2
    TransitionModality = self.calculateTransitionModality(peak_intensity_matrix)
    return np_mean(TransitionModality)

# Area Ratio(4)
  def calculateArea2SumRatioCV(self, chrom_data_list): #8(sum補0) # Level 3
    Area2SumRatios = np_array([x[6] for x in chrom_data_list])
    mean_ratio = np_mean(Area2SumRatios, axis=0)
    with np_errstate(divide='ignore', invalid='ignore'):
      Area2SumRatioCV = np_std(Area2SumRatios, axis=0, ddof=1) / mean_ratio
    np_nan_to_num(Area2SumRatioCV, copy=False, nan=5)
    return np_where(Area2SumRatioCV < 5, Area2SumRatioCV, 5) # Set the upperbound value to 5
  def calculatePeakGroupRatioCorr(self, area2sum_ratios):
    (light_ratios, heavy_ratios) = np_hsplit(area2sum_ratios, 2)
    if len(light_ratios) <= 2 or len(heavy_ratios) <= 2:
      corr = 0
    else:
      with np_errstate(divide='ignore', invalid='ignore'):
        corr = np_corrcoef(light_ratios, heavy_ratios)[0][1]
    np_nan_to_num(corr, copy=False, nan=0)
    return 0 if np_isnan(corr) else corr
  def calculateEachPairRatioConsistency(self, area2sum_ratios):
    (light_ratios, heavy_ratios) = np_hsplit(area2sum_ratios, 2)
    with np_errstate(divide='ignore', invalid='ignore'):
      PairRatioConsistency = np_abs((light_ratios - heavy_ratios)/heavy_ratios)
      np_nan_to_num(PairRatioConsistency, copy=False, nan=5)
      PairRatioConsistency = np_where(PairRatioConsistency < 5, PairRatioConsistency, 5)
    return PairRatioConsistency
  def calculatePairRatioConsistency(self, chrom_data_list):
    PairRatioConsistency = np_array([self.calculateEachPairRatioConsistency(x[6]) for x in chrom_data_list])
    return PairRatioConsistency
  def calculateMeanIsotopeRatioConsistency(self, chrom_data_list): #8(sum補0) #Level 3, each File has its own value
    Area2SumRatio = np_array([x[6] for x in chrom_data_list])
    mean = np_mean(Area2SumRatio, axis=0)
    with np_errstate(divide='ignore', invalid='ignore'):
      MeanIsotopeRatioConsistency = np_abs((Area2SumRatio - mean)/mean)
    np_nan_to_num(MeanIsotopeRatioConsistency, copy=False, nan=5)
    return np_where(MeanIsotopeRatioConsistency < 5, MeanIsotopeRatioConsistency, 5)  

# RT(1)
  def calculateMeanIsotopeRTConsistency(self, chrom_data_list): #1(sum補相同數值) # Level 3
    peakRT_crossAll = np_array([ (x[3][0] + x[3][1])/2 for x in chrom_data_list])
    mean = np_mean(peakRT_crossAll)
    with np_errstate(divide='ignore', invalid='ignore'):
      meanIsotopeRTconsistency = np_abs(peakRT_crossAll - mean) / mean
    np_nan_to_num(meanIsotopeRTconsistency, copy=False, nan=5)
    return np_where(meanIsotopeRTconsistency < 5, meanIsotopeRTconsistency, 5)

# TMSQE quality feature names
  def type1QualityTitle(self):
    return [
      'TransitionJaggedness_light',
      'TransitionSymmetry_light',
      'TransitionModality_light',
      'TransitionShift_light',
      'TransitionFWHM_light',
      'TransitionFWHM2base_light',
      'TransitionMaxBoundaryIntensityNormalized_light',
      'TransitionJaggedness_heavy',
      'TransitionSymmetry_heavy',
      'TransitionModality_heavy',
      'TransitionShift_heavy',
      'TransitionFWHM_heavy',
      'TransitionFWHM2base_heavy',
      'TransitionMaxBoundaryIntensityNormalized_heavy',
      'PairSimilarity',
      'PairShift',
      'PairFWHMConsistency'
    ]
  def type2QualityTitle(self):
    return [
      'IsotopeJaggedness_light',
      'IsotopeSymmetry_light',
      'IsotopeSimilarity_light',
      'IsotopeModality_light',
      'IsotopeShift_light',
      'IsotopeFWHM_light',
      'IsotopeFWHM2base_light',
      'IsotopeJaggedness_heavy',
      'IsotopeSymmetry_heavy',
      'IsotopeSimilarity_heavy',
      'IsotopeModality_heavy',
      'IsotopeShift_heavy',
      'IsotopeFWHM_heavy',
      'IsotopeFWHM2base_heavy',
      'PeakGroupRatioCorr',
      'PeakGroupJaggedness',
      'PeakGroupSymmetry',
      'PeakGroupSimilarity',
      'PeakGroupModality',
      'PeakGroupShift',
      'PeakGroupFWHM',
      'PeakGroupFWHM2base'
    ]
  def type3QualityTitle(self):
    return [
      "MeanIsotopeRatioConsistency_light",
      "MeanIsotopeFWHMConsistency_light",
      "Area2SumRatioCV_light",
      "MeanIsotopeRatioConsistency_heavy",
      "MeanIsotopeFWHMConsistency_heavy",
      "Area2SumRatioCV_heavy",
      "MeanIsotopeRTConsistency",
      "PairRatioConsistency"
    ]



from torch import tanh, no_grad, tensor, float32, load
from torch.nn import Linear, BatchNorm1d, Module
from torch.cuda import is_available as th_cuda_is_available
from numpy.linalg import norm as np_linalg_norm
from os.path import join as os_path_join, split as os_path_split

class Encoder(Module):
  def __init__(self, latent_dims, qc_level):
    super(Encoder, self).__init__()
    dims = []
    if qc_level == 1:
      dims = [17, 24, 8, latent_dims]
    elif qc_level == 2:
      dims = [22, 36, 12, latent_dims]
    elif qc_level == 3:
      dims = [8, 12, latent_dims]
    if qc_level == 3:
      self.linear1  = None
      self.linear2  = Linear(dims[0], dims[1])
      self.linear2_bn = BatchNorm1d(dims[1])
      self.linear3A = Linear(dims[1], dims[2])
      self.linear3B = Linear(dims[1], dims[2])
    else:
      self.linear1  = Linear(dims[0], dims[1])
      self.linear1_bn = BatchNorm1d(dims[1])
      self.linear2  = Linear(dims[1], dims[2])
      self.linear2_bn = BatchNorm1d(dims[2])
      self.linear3A = Linear(dims[2], dims[3])
      self.linear3B = Linear(dims[2], dims[3])
  def forward(self, x):
    if self.linear1 is not None:
      x = tanh(self.linear1(x))
      x = tanh(self.linear1_bn(x))
    x = tanh(self.linear2(x))
    x = tanh(self.linear2_bn(x))
    mu = self.linear3A(x)
    logvar = self.linear3B(x)
    return mu, logvar


class QualityEncoder(object):
  def __init__(self, device='auto', encoder_type1_path=None, encoder_type2_path=None, encoder_type3_path=None, encoder_dim = [2, 2, 2]):
    self.encoder_type1_path = encoder_type1_path
    self.encoder_type2_path = encoder_type2_path
    self.encoder_type3_path = encoder_type3_path
    if not self.encoder_type1_path:
      self.encoder_type1_path = os_path_join(os_path_split(__file__)[0], 'encoder', 'quality_encoder_type1.pickle')
    if not self.encoder_type2_path:
      self.encoder_type2_path = os_path_join(os_path_split(__file__)[0], 'encoder', 'quality_encoder_type2.pickle')
    if not self.encoder_type3_path:
      self.encoder_type3_path = os_path_join(os_path_split(__file__)[0], 'encoder', 'quality_encoder_type3.pickle')
    self.type1_encoder = Encoder(encoder_dim[0], qc_level=1)
    self.type2_encoder = Encoder(encoder_dim[1], qc_level=2)
    self.type3_encoder = Encoder(encoder_dim[2], qc_level=3)
    self.type1_quality_refs = [
      [0.0,  1.0, 0.0, 0.0, 0.050, 0.250, 0.0,    0.0,  1.0, 0.0, 0.0, 0.050, 0.250, 0.0,   1.0, 0.0, 0.0],
      [0.3,  0.4, 0.3, 0.3, 0.335, 0.475, 0.3,    0.3,  0.4, 0.3, 0.3, 0.335, 0.475, 0.3,   0.7, 0.3, 0.3],
      [1.0, -1.0, 1.0, 1.0, 1.000, 1.000, 5.0,    0.0,  1.0, 0.0, 0.0, 0.050, 0.250, 0.0,   0.0, 1.0, 1.0],
      [0.0,  1.0, 0.0, 0.0, 0.050, 0.250, 0.0,    1.0, -1.0, 1.0, 1.0, 1.000, 1.000, 5.0,   0.0, 1.0, 1.0],
      [1.0, -1.0, 1.0, 1.0, 1.000, 1.000, 5.0,    1.0, -1.0, 1.0, 1.0, 1.000, 1.000, 5.0,   0.0, 1.0, 1.0]
    ]
    self.type2_quality_refs = [
      [0.0,  1.0,  1.0, 0.0, 0.0, 0.200, 0.200,   0.0,  1.0,  1.0, 0.0, 0.0, 0.200, 0.200,   1.0, 0.0,  1.0,  1.0, 0.0, 0.0, 0.200, 0.200],
      [0.2,  0.6,  0.6, 0.2, 0.2, 0.360, 0.360,   0.2,  0.6,  0.6, 0.2, 0.2, 0.360, 0.360,   0.6, 0.2,  0.6,  0.6, 0.2, 0.2, 0.360, 0.360],
      [1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000,   0.0,  1.0,  1.0, 0.0, 0.0, 0.050, 0.250,  -1.0, 1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000], # good heavy only; worst others
      [0.0,  1.0,  1.0, 0.0, 0.0, 0.050, 0.250,   1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000,  -1.0, 1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000], # bad heavy only
      [1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000,   1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000,  -1.0, 1.0, -1.0, -1.0, 1.0, 1.0, 1.000, 1.000]
    ]
    self.type3_quality_refs = [
      [0.0,  0.0,  0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
      [1.5,  0.3,  0.9, 1.5, 0.3, 0.9, 0.3, 0.3],
      [5.0,  1.0,  3.0, 0.0, 0.0, 0.0, 1.0, 1.0],
      [0.0,  0.0,  0.0, 5.0, 1.0, 3.0, 1.0, 1.0],
      [5.0,  1.0,  3.0, 5.0, 1.0, 3.0, 1.0, 1.0]
    ]
    if device == 'auto':
      self.device = "cuda" if th_cuda_is_available() else "cpu"
    else:
      self.device = device
    print('device: ' + self.device)
    self.load_encoder()
    self._update_reference_latents()
  
  def __call__(self, quality_vectors, normalize=True):
    with no_grad():
      qv = tensor(quality_vectors, dtype=float32).to(self.device)
      dim = qv.shape[1]
      if dim == 17:
        latent = self.type1_encoder(qv)[0].cpu().detach().numpy()
        score = self.score_type1_latent(latent)
        type = 1
      elif dim == 22:
        latent = self.type2_encoder(qv)[0].cpu().detach().numpy()
        score = self.score_type2_latent(latent)
        type = 2
      elif dim == 8:
        latent = self.type3_encoder(qv)[0].cpu().detach().numpy()
        score = self.score_type3_latent(latent)
        type = 3
      else:
        raise 'Unkonw dimension. Valid dimensions are 17 for type 1 quality, 22 for type 2 quality, and 8 for type 3 quality.'
    if normalize:
      return self.normalize_score(type, score), latent
    else:
      return score, latent

  def load_encoder(self):
    self.type1_encoder.load_state_dict(load(self.encoder_type1_path, map_location=self.device, weights_only=True))
    self.type2_encoder.load_state_dict(load(self.encoder_type2_path, map_location=self.device, weights_only=True))
    self.type3_encoder.load_state_dict(load(self.encoder_type3_path, map_location=self.device, weights_only=True))
    self.type1_encoder.to(self.device)
    self.type2_encoder.to(self.device)
    self.type3_encoder.to(self.device)
    self.type1_encoder.eval()
    self.type2_encoder.eval()
    self.type3_encoder.eval()
    self._update_reference_latents()

  def _update_reference_latents(self):
    self.type1_latent_points = self.type1_encoder(tensor(self.type1_quality_refs).to(self.device))[0].to('cpu').detach().numpy()
    self.type2_latent_points = self.type2_encoder(tensor(self.type2_quality_refs).to(self.device))[0].to('cpu').detach().numpy()
    self.type3_latent_points = self.type3_encoder(tensor(self.type3_quality_refs).to(self.device))[0].to('cpu').detach().numpy()
    self._score_range = [
      dict(max= self.score_type1_latent([self.type1_latent_points[0]])[0], min=self.score_type1_latent([self.type1_latent_points[4]])[0]),
      dict(max= self.score_type2_latent([self.type2_latent_points[0]])[0], min=self.score_type2_latent([self.type2_latent_points[4]])[0]),
      dict(max= self.score_type3_latent([self.type3_latent_points[0]])[0], min=self.score_type3_latent([self.type3_latent_points[4]])[0])
    ]
  def encode_quality(self, quality_vectors):
    qv = tensor(quality_vectors, dtype=float32).to(self.device)
    dim = qv.shape[1]
    if dim == 17:
      return self.type1_encoder(tensor(qv, dtype=float32).to(self.device))[0].detach().numpy()
    elif dim == 22:
      return self.type2_encoder(tensor(qv, dtype=float32).to(self.device))[0].detach().numpy()
    elif dim == 8:
      return self.type3_encoder(tensor(qv, dtype=float32).to(self.device))[0].detach().numpy()
    else:
      raise 'Unkonw dimension. Valid dimensions are 17 for type 1 quality, 22 for type 2 quality, and 8 for type 3 quality.'

  def score_type1_latent(self, type1_latents):
    return self.score_func(type1_latents, self.type1_latent_points)
  def score_type2_latent(self, type2_latents):
    return self.score_func(type2_latents, self.type2_latent_points)
  def score_type3_latent(self, type3_latents):
    return self.score_func(type3_latents, self.type3_latent_points)

  def normalize_score(self, type, score):
    if type <= 0 or type >=4:
      return
    min = self._score_range[type - 1]['min']
    max = self._score_range[type - 1]['max']
    return -10 + 20 * ((score - min)/(max - min))

  def score_func(self, latent_points, ref_latent_points):
    # ref_latent_points = self._ref_latent_points
    dist_max = np_linalg_norm(ref_latent_points[0] - ref_latent_points[4])
    score = 2 * (1 - (self.dist_func(latent_points, ref_latent_points[0])/dist_max)**0.5)
    score = score + 1 * (1 - (self.dist_func(latent_points, ref_latent_points[1])/dist_max)**0.5)
    score = score + 1 * (1 - (self.dist_func(latent_points, ref_latent_points[2])/dist_max)**0.5)
    score = score - 1 * (1 - (self.dist_func(latent_points, ref_latent_points[3])/dist_max)**0.5)
    score = score - 2 * (1 - (self.dist_func(latent_points, ref_latent_points[4])/dist_max)**0.5)
    return score

  def dist_func(self, a, b):
    return np_linalg_norm(a - b, axis=1)



from os import environ as os_environ
os_environ["OMP_NUM_THREADS"] = "1"
from argparse import ArgumentParser
from MsTargetPeaker import PeakQualityReport


def main():
  parser = ArgumentParser()
  parser.add_argument("chromatogram_tsv", type=str, help="The chromatogram TSV file path")
  parser.add_argument("peak_boundary_csv", type=str, help="The Peak Boundary CSV file path")
  parser.add_argument("output_folder", type=str, help="Output folder")
  parser.add_argument('--internal_standard_type', '-s',type=str, default='heavy', choices=['heavy', 'light'] , help="Set the internal standards to heavy or light ions. (default: heavy)") 
  parser.add_argument('--top_n_fragment', '-n', type=int, default=5, help="Automatically select top N transition ion for quality reporting. (default: 5)")
  parser.add_argument('--group_csv', '-g', type=str, default=None, help="The CSV file containing the ReplicateName column indicating the groups/batch of samples")
  parser.add_argument('--output_chromatogram_pdf', '-pdf', action='store_true', help="Set for making chromatogram plots in a chromatogram_plots.pdf file. (default: unset)")
  parser.add_argument('--chromatogram_dpi', '-dpi', type=str, default=200, help="The dpi of chromatogram plots. Only works when --output_chromatogram_pdf is set. (default: 200)")
  parser.add_argument('--chromatogram_nrow', '-nrow', type=int, default=6, help="The number of chromatograms per row in one pdf page. Only works when --output_chromatogram_pdf is set. (default: 6)")
  parser.add_argument('--chromatogram_ncol', '-ncol', type=int, default=6, help="The number of chromatograms per columne in one pdf page. Only works when --output_chromatogram_pdf is set. (default: 6)")
  parser.add_argument('--chromatogram_fig_w', '-figw', type=int, default=None, help="The figure width in inches. Only works when --output_chromatogram_pdf is set. (default: nrow * 7)")
  parser.add_argument('--chromatogram_fig_h', '-figh', type=int, default=None, help="The figure height in inches. Only works when --output_chromatogram_pdf is set. (default: ncol * 5)")
  parser.add_argument('--output_mixed_mol', '-mix', action='store_true', help="If set, chromatogram data for each molecule will be mixed in one pdf page. (default: unset)")
  parser.add_argument('--reorder_by_group', '-r', action='store_true', help="If set, chromatogram data for each molecule will be mixed in one pdf page. (default: unset)")
  args = parser.parse_args()
  print("Input Parameters:")
  for arg in vars(args):
    print(arg, '=',getattr(args, arg))
  print('#########################################################')
  pqr = PeakQualityReport(args.chromatogram_tsv, args.peak_boundary_csv, args.output_folder, internal_standard_type=args.internal_standard_type, top_n_transitions=args.top_n_fragment, group_csv=args.group_csv)
  pqr.run()
  if args.output_chromatogram_pdf:
    pqr.output_chrom_plot(ncol=args.chromatogram_ncol, nrow=args.chromatogram_nrow, dpi=args.chromatogram_dpi, figW=args.chromatogram_fig_w, figH=args.chromatogram_fig_h, mol_separation=not args.output_mixed_mol, reorder=args.reorder_by_group)
  print('Finished')

if __name__ == '__main__':
  main()




from pandas import read_csv
from numpy import (arange as np_arange, hsplit as np_hsplit, array as np_array, vstack as np_vstack, sqrt as np_sqrt, where as np_where, quantile as np_quantile, float32 as np_float32, mean as np_mean, average as np_average, power as np_power, sum as np_sum, absolute as np_absolute, unique as np_unique, zeros as np_zeros, max as np_max, argmax as np_argmax, log as np_log, count_nonzero as np_count_nonzero, median as np_median, argsort as np_argsort, abs as np_abs, sort as np_sort, flip as np_flip, vectorize as np_vectorize, zeros_like as np_zeros_like,
                  sign as np_sign, diff as np_diff, around as np_around, clip as np_clip, all as np_all, amin as np_amin, amax as np_amax, apply_along_axis as np_apply_along_axis)
from scipy.stats import (percentileofscore as scipy_percentileofscore, gmean as scipy_gmean)
from numpy.random import shuffle as np_random_shuffle
from torch import no_grad as th_no_grad, Tensor as th_Tensor, tensor as th_tensor, as_tensor as th_as_tensor, load as th_load, FloatTensor as th_FloatTensor, float32 as th_float32
from torch.cuda import is_available as th_cuda_is_available
from torch.nn import Module, Sequential, Conv1d, MaxPool1d, ReLU, Flatten, Tanh, Linear
from gymnasium.spaces import Box as gym_spaces_Box
from .tmsqe import PeakQualityEval
from .tmsqe import ChromatogramDB
from .mcts.MCTS_DPW import MCTS_DPW
from scipy.stats import (norm as scipy_norm, gaussian_kde as scipy_gaussian_kde)
from gc import collect as gc_collect
from os.path import join as os_path_join, split as os_path_split

def _shuffle_ions(obs_mx):
  reshaped_array = obs_mx[2:,:].reshape(-1, 2, obs_mx.shape[1])
  np_random_shuffle(reshaped_array)
  shuffled_obs_mx = reshaped_array.reshape(-1, obs_mx.shape[1])
  return np_vstack([obs_mx[0:2,:], shuffled_obs_mx])

def _chrom_to_current_obs(chrom_data):
  (start, end) = chrom_data[3]
  time = chrom_data[10]
  intensity = chrom_data[11]
  transitions = chrom_data[2][0]
  transition_len = len(transitions)
  random_pick_transition_idx = np_arange(transition_len)
  np_random_shuffle(random_pick_transition_idx)
  shuffled_idx = random_pick_transition_idx[0:20]
  transitions = [transitions[transition_idx] for transition_idx in shuffled_idx]
  (light_inst, heavy_inst) = np_hsplit(intensity, 2)
  light_inst = _normalize_intensity(light_inst[:, shuffled_idx])
  heavy_inst = _normalize_intensity(heavy_inst[:, shuffled_idx])
  binary_mask = 1*(1*(time >= start) & (time <= end))
  step_indicator = np_array([1]*len(time))
  light_heavy_pairs = np_vstack(list(zip(light_inst.T, heavy_inst.T)))
  max_channels = 42
  to_be_padded = max_channels - 2 - light_heavy_pairs.shape[0]
  obs_mx = np_vstack([step_indicator, binary_mask, light_heavy_pairs, np_zeros((to_be_padded, 1024))])
  return _shuffle_ions(obs_mx)

def _reset_obs(obs_mx, chrom_data):
  obs_mx[0,:] = 1
  (start, end) = chrom_data[3]
  time = chrom_data[10]
  obs_mx[1,:] = 1*(1*(time >= start) & (time <= end))
  return _shuffle_ions(obs_mx)

def _normalize_intensity(intensity_values):
  intensity_df = np_sqrt(np_where(intensity_values >= 1, intensity_values, 1)) #2nd_env57
  max = intensity_df.max().max()
  if max == 0.0:
    return np_zeros_like(intensity_df)
  else:
    min = np_quantile(intensity_df, 0.3)
    if max - min == 0:
      return np_zeros_like(intensity_df)
    normalized = 1 * ((intensity_df - min)/(max-min))
    return np_where(normalized >= 0, normalized, 0)

def _get_quantile_factor(chrom_data, top_n = 1):
  intensity = chrom_data[9]
  peak_intensity = chrom_data[4]
  if peak_intensity.shape[0] == 0:
    return 0, 0
  heavy_peak_intensity = np_hsplit(peak_intensity, 2)[1]
  heavy_intensity = np_hsplit(intensity, 2)[1]
  heavy_unique_intensity = np_unique(heavy_intensity)
  if top_n == 1:
    heavy_peak_intensity_max = np_max(heavy_peak_intensity)
    heavy_peakheight_quantile = np_count_nonzero(heavy_unique_intensity <= heavy_peak_intensity_max) / heavy_unique_intensity.size
  else:
    heavy_peak_intensity_max = np_max(heavy_peak_intensity, axis=0)
    heavy_peakheight_quantiles = scipy_percentileofscore(heavy_unique_intensity, heavy_peak_intensity_max)/100
    heavy_sorted_quantiles = -np_sort(-heavy_peakheight_quantiles)
    heavy_peakheight_quantile = np_mean(heavy_sorted_quantiles[:top_n])
  peak_height_quantile = heavy_peakheight_quantile
  intensity_quantile_factor = 1 - ((1-peak_height_quantile)**0.4)
  return peak_height_quantile, intensity_quantile_factor

def _get_pbar_factor(chrom_data, pbar_heavy_weight=2/3, use_weighted_pbar=True):
  (start, end) = chrom_data[3]
  peak_intensity = chrom_data[4]
  area = chrom_data[7]
  intensity = chrom_data[9]
  if peak_intensity.shape[0] < 1:
    return 0, 1.2e-08, 0, np_array([0.0, 1.0, 2.0]), np_array([1.0, 1.0, 1.0])
  (peak_intensity_light, peak_intensity_heavy) = np_hsplit(peak_intensity, 2)
  (area_light, area_heavy) = np_hsplit(area, 2)
  peak_intensity_max_all = peak_intensity_heavy.max(axis=0)
  picked_area_heavy = (end - start) * peak_intensity_max_all # heavy intensity
  picked_area_heavy[picked_area_heavy < 0.01] = 0.01
  pbar_heavy = area_heavy/picked_area_heavy
  picked_area_light = (end - start) * peak_intensity_light.max(axis=0) # light intensity
  picked_area_light[picked_area_light < 0.01] = 0.01
  pbar_light = area_light/picked_area_light
  ordered_transition_list = np_flip(np_argsort(peak_intensity_max_all)[-3:])
  peak_intensity_max_all = np_where(peak_intensity_max_all > 0, peak_intensity_max_all, 1e-8)
  intensity_ratios = peak_intensity_max_all/peak_intensity_max_all.max()
  intensity_ratios = intensity_ratios[ordered_transition_list]
  if use_weighted_pbar:
    weights = np_power(intensity_ratios, 2)
    weighted_pbar = (pbar_heavy_weight)*pbar_heavy + (1 - pbar_heavy_weight)*pbar_light
    pbar = np_average(weighted_pbar[ordered_transition_list], weights=weights)
  else:
    weighted_pbar = (pbar_heavy_weight)*pbar_heavy + (1 - pbar_heavy_weight)*pbar_light
    pbar = np_mean(weighted_pbar[ordered_transition_list])
  means = [0.3, 0.4, 0.5]
  sigmas = [0.05, 0.06, 0.07]
  pbar_factor = sum(scipy_norm.pdf(pbar, mu, sigmas[i]) for i, mu in enumerate(means))/10.33728076626613
  max_peak_intensity_ion_index = peak_intensity_max_all.argmax()
  intensity_heavy = np_hsplit(intensity, 2)[1]
  peak_mean = peak_intensity_heavy[:, max_peak_intensity_ion_index].mean()
  standard_intensity = intensity_heavy.mean(axis=0)
  sorted_mean = -np_sort(-standard_intensity)
  top_five_global_mean = sorted_mean[:3].mean()
  isSignal = 1 if peak_mean > 0.5*top_five_global_mean else 0
  return pbar, pbar_factor, isSignal, ordered_transition_list, intensity_ratios

def _get_pair_ratio_consistency_factor(ms_peak_qc, chrom_data, pair_ratio_factor):
  pair_ratio_consistency = ms_peak_qc['qualityEval'].calculateEachPairRatioConsistency(chrom_data[6])
  pair_ratio_consistency_median = np_median(pair_ratio_consistency)
  adjusted_pair_ratio_consistency = pair_ratio_consistency_median * pair_ratio_factor
  pair_ratio_consistency_factor = 0.0 if (adjusted_pair_ratio_consistency) >= 2 else ((2 - adjusted_pair_ratio_consistency)**2)/4
  return pair_ratio_consistency_median, pair_ratio_consistency_factor

def _each_transition_modality(signal_series):
    max_intensity = np_max(signal_series)
    if max_intensity == 0: # No intensity
      return 100.0
    flatness_factor = 0.0 if max_intensity <= 1000 else  0.05
    rise = []
    fall = []
    total = 0.0
    sign = np_sign(signal_series[0])
    rise_fall_list = []
    for _, intensity_diff in enumerate(signal_series):
      if np_sign(intensity_diff) == sign:
        total += intensity_diff
      else:
        if np_abs(total) > (flatness_factor * max_intensity):
          if sign > 0:
            rise.append(total)
            rise_fall_list.append(total)
          else:
            fall.append(total)
            rise_fall_list.append(total)
        total = intensity_diff
        sign = np_sign(intensity_diff)
    if np_sign(intensity_diff) == sign:
      if sign > 0:
        rise.append(total)
        rise_fall_list.append(total)
      else:
        fall.append(total)
        rise_fall_list.append(total)
    if len(rise) == 0 or len(fall) == 0: #no peaks in the region
      return 100.0
    sorted_rise_idx = np_argsort(rise)
    sorted_fall_idx = np_argsort(fall)
    rise_dip = rise[sorted_rise_idx[-2]] if len(rise) >= 2 else 0
    fall_dip = fall[sorted_fall_idx[1]] if len(fall) >= 2 else 0
    main_rise = rise[sorted_rise_idx[-1]]
    main_dip = fall[sorted_fall_idx[0]]
    main_rise_idx = rise_fall_list.index(main_rise)
    main_dip_idx = rise_fall_list.index(main_dip)
    if main_rise == 0 or main_dip == 0: # No main peaks in the region
      return 100.0
    elif main_rise_idx > main_dip_idx: # No main peaks in the region (continuous decreasing or increasing signals
      return 100.0
    max_dip = max(np_abs(fall_dip),  rise_dip)
    if max_dip == 0 and np_sign(signal_series[0]) < 0:  # (The first sign is negative and no dip => the velley signal pattern in the region
      return 100.0
    rise_fall_ratio = np_absolute(np_log(np_absolute(main_rise/main_dip)))
    modality = (max_dip / max_intensity) + (rise_fall_ratio/3)
    return 1.0 if modality > 1.0 else 0.0 if modality < 0.0 else modality

def _get_peak_modality_factor(chrom_data, ordered_transition_list, intensity_ratios, isHeavyFirst):
  peak_intensity = chrom_data[4]
  if peak_intensity.shape[0] < 2:
    return 100.0, 0.01
  (_, peak_intensity_heavy) = np_hsplit(peak_intensity, 2)
  peak_diff_mx = np_diff(peak_intensity_heavy, axis=0)
  modality_heavy = np_apply_along_axis(_each_transition_modality, 0, peak_diff_mx)
  modality_heavy = modality_heavy[ordered_transition_list]
  weights = np_power(intensity_ratios, 2)
  modality = np_average(modality_heavy, weights=weights)
  if modality >= 100: #Invalid peaks
    modality_factor = 0.01
  elif modality > 1:
    modality_factor = 0.1
  else:
    if modality <= 1:
      adjusted_modality = modality * 0.3
      modality_factor = (1 - adjusted_modality**0.4) if adjusted_modality < 1 else 0.0
    else:
      modality_factor = 0.1
    modality_factor = 0.1 if modality_factor < 0.1 else modality_factor
  if isHeavyFirst and modality != 100:
    modality_factor = 0.3 if modality_factor < 0.3 else modality_factor
  return modality, modality_factor

def _get_peak_boundary_consensus_factor(chrom_data, peak_kde):
  mol = chrom_data[0]
  start_kde_factor = -1
  end_kde_factor = -1
  kde_factor = 1
  isHeavyFirst = False
  if peak_kde and mol in peak_kde:
    if peak_kde[mol] == True:
      isHeavyFirst = True
    elif peak_kde[mol] == False:
      pass
    else:
      (peak_start, peak_end) = chrom_data[3]
      start_kde_func = peak_kde[mol]['start']
      end_kde_func = peak_kde[mol]['end']
      start_kde_factor = start_kde_func(peak_start)[0]/peak_kde[mol]['start_max']
      end_kde_factor = end_kde_func(peak_end)[0]/peak_kde[mol]['end_max']
      if start_kde_factor < 1e-5:
        start_kde_factor2 = 0.0
      else:
        start_kde_factor2 = start_kde_factor if start_kde_factor > 0.1 else 0.1
      if end_kde_factor < 1e-5:
        end_kde_factor2 = 0.0
      else:
        end_kde_factor2 = end_kde_factor if end_kde_factor > 0.1 else 0.1
      if start_kde_factor2 == 0 or end_kde_factor2 == 0:
        kde_factor = 0.01
      else:
        kde_factor = (start_kde_factor2 + end_kde_factor2)/2
  return kde_factor, start_kde_factor, end_kde_factor, isHeavyFirst

def _serialize_param(pair_ratio_factor, intensity_power, quality_power, top_n_ion):
  return f'{quality_power}_{top_n_ion}_{pair_ratio_factor}_{intensity_power}'

def _get_quality_reward(ms_peak_qc, chrom_data, isHeavyFirst):
  type_1_reward, type_2_reward = ms_peak_qc['qualityEval'].getOverallScore(chrom_data)
  raw_reward = type_1_reward + type_2_reward
  quality_reward = raw_reward / 17.605
  if quality_reward >= 0 and quality_reward < 0.1:
    quality_reward = 0.1
  elif quality_reward < 0:
    quality_reward = 0.1 if isHeavyFirst else 0.1 * ((1 + quality_reward)**2)
  quality_reward = 0 if quality_reward < 0 else quality_reward
  return quality_reward, type_1_reward, type_2_reward, raw_reward

def _calculate_chrom_score(ms_peak_qc, current_chrom, pair_ratio_factor=0.3, pbar_heavy_weight=2/3, intensity_power=1, quality_power=2.0, peak_kde=None, top_n_ion=1, use_weighted_pbar=True):
  kde_factor, start_kde_factor, end_kde_factor, isHeavyFirst = _get_peak_boundary_consensus_factor(current_chrom, peak_kde)
  if isHeavyFirst:          # Enter Heavy-First mode if no valid peaks were found in the target.
    intensity_power = 10    # We used a stronger factor to encourage selections of peak regions with strong heavy signals.
    pair_ratio_factor = 0   # In this case, we do not need to consider the pair ratio consistency.
  pbar, pbar_factor, isSignal, ordered_transition_list, intensity_ratios = _get_pbar_factor(current_chrom, pbar_heavy_weight=pbar_heavy_weight, use_weighted_pbar=use_weighted_pbar)
  pair_ratio_consistency_median, pair_ratio_consistency_factor = _get_pair_ratio_consistency_factor(ms_peak_qc, current_chrom, pair_ratio_factor)
  peak_modality, modality_factor = _get_peak_modality_factor(current_chrom, ordered_transition_list, intensity_ratios, isHeavyFirst)
  intensity_quantile, intensity_quantile_factor  = _get_quantile_factor(current_chrom, top_n=top_n_ion)
  quality_reward, type_1_reward, type_2_reward, raw_reward = _get_quality_reward(ms_peak_qc, current_chrom, isHeavyFirst)
  reward = (quality_reward**quality_power) * isSignal * pbar_factor * (intensity_quantile_factor**intensity_power) * pair_ratio_consistency_factor * modality_factor * kde_factor
  final_reward = -1.0 if reward <= 0.0 else (reward**1.5)*10
  param = _serialize_param(pair_ratio_factor, intensity_power, quality_power, top_n_ion)
  return dict(raw=(type_1_reward, type_2_reward),
              raw_reward=raw_reward,
              reward_score=quality_reward,
              isSignal=isSignal,
              PBAR=pbar,
              PBAR_factor=pbar_factor,
              pair_ratio_consistency_median=pair_ratio_consistency_median,
              pair_ratio_consistency_factor=pair_ratio_consistency_factor,
              peak_modality=peak_modality,
              modality_factor=modality_factor,
              intensity_quantile=intensity_quantile,
              intensity_quantile_factor=intensity_quantile_factor, 
              peak_start_factor=start_kde_factor,
              peak_end_factor=end_kde_factor,
              peak_location_factor=kde_factor,
              reward=reward,
              final_reward=final_reward,param=param)

def _process_peak_idx(s, e):
  if s > e:
    s, e = e, s
  elif s == e:
    s = s - 1
    e = e + 1
  s = max(0, min(1023, s))
  e = max(0, min(1023, e))
  if s >= e:
    if e < 1023:
      e += 1
    elif s > 0:
      s -= 1
  if s > e:
    s, e = e, s
  return s, e

def _step(current_obs, ms_peak_qc, current_chrom, current_step, action, history=None, max_step=5, pair_ratio_factor=0.5,
          pbar_heavy_weight=2/3, intensity_power=1, quality_power = 2.0,
          peak_kde=None, top_n_ion=1, use_weighted_pbar=True):
  index_offset_factor = 200
  rt_time = current_chrom[10]
  start_index_offset = int(np_around(action[0] * index_offset_factor))
  end_index_offset = int(np_around(action[1] * index_offset_factor))
  current_obs = current_obs.copy()
  try:
    peak_region = np_where(current_obs[1, :] >= 1)[0]
    if len(peak_region) == 0:
      (s, e) = ms_peak_qc['chromDB'].getRandomBoundary(current_chrom)
      current_chrom = ms_peak_qc['chromDB'].updateChromData(current_chrom, start=s, end=e)
      current_obs = _chrom_to_current_obs(current_chrom)
      return _step(current_obs, ms_peak_qc, current_chrom, current_step, action, history=history, max_step=max_step, 
                   pair_ratio_factor=pair_ratio_factor,
                   pbar_heavy_weight=pbar_heavy_weight, intensity_power=intensity_power, quality_power=quality_power,
                   peak_kde=peak_kde, top_n_ion=top_n_ion, use_weighted_pbar=use_weighted_pbar)
    elif len(peak_region) == 1:
      (current_start_index, current_end_index) = peak_region[0], peak_region[0]
    else:
      (current_start_index, current_end_index) = peak_region[[0, -1]]
  except Exception as e:
    print(np_where(current_obs[1, :] >= 1))
    raise e
  new_start_idx = current_start_index + start_index_offset
  new_end_idx = current_end_index + end_index_offset
  done = False
  (new_start_idx, new_end_idx) = _process_peak_idx(new_start_idx, new_end_idx)
  current_obs[0,:] = 1 - ((current_step + 1) * (1/max_step))
  current_obs[1,:] = 0
  current_obs[1,new_start_idx:new_end_idx+1] = 1
  (new_start_time, new_end_time)= rt_time[[new_start_idx, new_end_idx]]
  current_chrom = ms_peak_qc['chromDB'].updateChromData(current_chrom, start=new_start_time, end=new_end_time)
  current_step = current_step + 1
  is_stopping = current_step >= max_step
  score_obj = _calculate_chrom_score(ms_peak_qc, current_chrom, pair_ratio_factor=pair_ratio_factor, 
                                     pbar_heavy_weight=pbar_heavy_weight, 
                                     top_n_ion=top_n_ion,
                                     use_weighted_pbar=use_weighted_pbar,
                                     intensity_power=intensity_power, quality_power=quality_power, peak_kde=peak_kde)
  if history is not None:
    history.append(dict(
      sampleName = current_chrom[1],
      peptide=current_chrom[0],
      peak_start = current_chrom[3][0],
      peak_end = current_chrom[3][1],
      type_1_reward = score_obj['raw'][0],
      type_2_reward = score_obj['raw'][1],
      final_reward = score_obj['final_reward'],
      PBAR = score_obj['PBAR'],
      PBAR_factor = score_obj['PBAR_factor'],
      pair_ratio_consistency_median = score_obj['pair_ratio_consistency_median'],
      pair_ratio_consistency_factor = score_obj['pair_ratio_consistency_factor'],
      peak_modality = score_obj['peak_modality'],
      modality_factor = score_obj['modality_factor'],
      intensity_quantile = score_obj['intensity_quantile'],
      intensity_quantile_factor = score_obj['intensity_quantile_factor'],
      peak_start_factor = score_obj['peak_start_factor'],
      peak_end_factor = score_obj['peak_end_factor'],
      peak_location_factor = score_obj['peak_location_factor'],
      notes = score_obj['param']
    ))
  if is_stopping:
    done = True
    reward = score_obj['reward']
  current_obs = _shuffle_ions(current_obs)
  info = dict(chrom = current_chrom, score=score_obj)
  return current_obs, score_obj['reward'], done, info

class MSPeakFeatureExtractorNet(Module):
  def __init__(self, observation_space: gym_spaces_Box, features_dim=128, latent_dim=512):
    super().__init__()
    assert features_dim > 0
    self._observation_space = observation_space
    self._features_dim = features_dim
    self.convs = Sequential(
      Conv1d(observation_space.shape[0], 64, kernel_size = 3, stride=2),
      MaxPool1d(kernel_size=2),
      ReLU(),
      Conv1d(64, 32, kernel_size=3, stride=2),
      MaxPool1d(kernel_size=2),
      ReLU(),
      Flatten()
    )
    with th_no_grad():
      n_flatten = self.convs(th_as_tensor(observation_space.sample()[None]).float()).shape[1]
    self.fc = Sequential(
      Linear(n_flatten, latent_dim),
      Tanh(),
      Linear(latent_dim, features_dim),
      Tanh()
    )
  @property
  def features_dim(self) -> int:
    return self._features_dim
  def forward(self, observations: th_Tensor):
    return self.fc(self.convs(observations))

class MlpExtractor(Module):
  def __init__(self):
    super().__init__()
    self.policy_net = Sequential(
      Linear(128, 32, bias=True),
      Tanh(),
      Linear(32, 32, bias=True),
      Tanh()
    )
    self.value_net = Sequential(
      Linear(128, 32, bias=True),
      Tanh(),
      Linear(32, 32, bias=True),
      Tanh()
    )
  def forward(self, features: th_Tensor):
    latent_pi = self.policy_net(features)
    latent_vf = self.value_net(features)
    return latent_pi, latent_vf

class MsTargetPeakerPolicy(Module):
  def __init__(self, observation_space: gym_spaces_Box, features_dim=128, latent_dim=512):
    super().__init__()
    self.features_extractor = MSPeakFeatureExtractorNet(observation_space, features_dim, latent_dim)
    self.mlp_extractor = MlpExtractor()
    self.action_net = Linear(32, 2, bias=True)
    self.value_net = Linear(32, 1, bias=True)
  def forward(self,  obs: th_Tensor):
    features = self.features_extractor(obs)
    latent_pi, latent_vf = self.mlp_extractor(features)
    actions = self.action_net(latent_pi)
    values = self.value_net(latent_vf).detach()
    return actions, values

class MsTargetPeakerEnv():
  def __init__(self, chrom_tsv, policy_path=None, picked_peak_csv=None,
               max_step=5, device=None, 
               internal_standard_type='heavy',
               threshold_for_kde=2, 
               pair_ratio_factor=0.5,
               pbar_heavy_weight=2/3,
               intensity_power=1.0,
               quality_power=2.0,
               top_n_ion=1,
               use_weighted_pbar=True,
               use_kde=None,
               use_ref_if_no_kde=False):
    self.internal_standard_type = internal_standard_type
    self.max_time_point = 1024
    self.nFragment = 20
    self.max_step = max_step
    if device is None:
      self.device = 'cuda' if th_cuda_is_available() else 'cpu'
    else:
      self.device = device
    self.policy = MsTargetPeakerPolicy(gym_spaces_Box(low=0, high=1, shape=(42, 1024), dtype=np_float32), 128, 512)
    if policy_path is None:
      policy_path = os_path_join(os_path_split(__file__)[0], 'policy', 'MsTargetPeaker_env66_40-2.state_dict.pth')
    self.load_policy_state(policy_path)
    self.picked_peak_csv = picked_peak_csv
    self.peak_qc = dict(
      chromDB = ChromatogramDB(chrom_tsv, picked_peak_csv, internal_standard_type=internal_standard_type),
      qualityEval = PeakQualityEval(device=self.device)
    )
    self.pair_ratio_factor = pair_ratio_factor
    self.pbar_heavy_weight = pbar_heavy_weight
    self.intensity_power = intensity_power
    self.picked_peak_for_kde = use_kde or None
    self.peak_kde = None
    self.top_n_ion = top_n_ion
    self.quality_power = quality_power
    self.use_weighted_pbar = use_weighted_pbar
    print(f'Used picked peaks for kde: {use_kde}')
    if self.picked_peak_for_kde:
      kde_peak_df = read_csv(self.picked_peak_for_kde)
      kde_peak_df.columns = kde_peak_df.columns.str.replace(' ', '')
      self.peak_kde = {}
      print(f"Threshold for Peak Location KDE: {threshold_for_kde}.")
      print(f"Enter reference-first mode when no qualified peaks to form a KDE function: {use_ref_if_no_kde}.") 
      for pep, data in kde_peak_df.groupby('PeptideModifiedSequence'):
        if 'FinalReward' in data:
          valid_data = data[data['FinalReward'] >= threshold_for_kde]
        else:
          valid_data = data
        if len(valid_data) < 3:
          print(f'Cannot establish probability density functions of peak locations for moledcule {pep} because the number of peak groups are less than 3. Skipped.')
          if use_ref_if_no_kde:
            self.peak_kde[pep] = True # Reference-first mode to consider peaks mainly based on the reference peak height. (since no valid peaks can be used as references for finding peak locations in the last step of the rescue stage.)
          else:
            self.peak_kde[pep] = False # Use the provided search parameters to find peak locations.
          continue
        start_data = valid_data.sort_values(by='MinStartTime')
        x1 = start_data['MinStartTime']
        x1_weights= start_data['FinalReward'] if 'FinalReward' in start_data else None
        try:
          x1_kde = scipy_gaussian_kde(x1, weights=x1_weights)
        except:
          x1_kde = np_vectorize(lambda x: 1 if abs(x - np_mean(x1)) < 0.01 else 0)
        start_max = np_max(x1_kde(x1))
        end_data = valid_data.sort_values("MaxEndTime")
        x2 = end_data['MaxEndTime']
        x2_weights = end_data['FinalReward'] if 'FinalReward' in end_data else None
        try:
          x2_kde = scipy_gaussian_kde(x2, weights=x2_weights)
        except:
          x2_kde = np_vectorize(lambda x: 1 if abs(x - np_mean(x2)) < 0.01 else 0)
        end_max = np_max(x2_kde(x2))
        self.peak_kde[pep] = {
          'start': x1_kde,
          'end': x2_kde,
          'start_max': start_max,
          'end_max': end_max,
          'start_rt': x1.iloc[np_argmax(x1_kde(x1))],
          'end_rt': x2.iloc[np_argmax(x2_kde(x2))]
        }
      print(f'Got peak location distributions using kernel-density estimations from peaks in file {self.picked_peak_for_kde}.')
    print(f'Env option: pair_ratio_factor: {self.pair_ratio_factor}; pbar_heavy_weight: {self.pbar_heavy_weight}; intensity_power: {self.intensity_power}; quality_power: {self.quality_power}; use_kde: {"True" if self.peak_kde else "False"}')
    
  def load_policy_state(self, policy_path):
    self.policy.load_state_dict(th_load(policy_path, weights_only=True))
    self.policy.to(self.device)
    self.policy.eval()
    print(f'Successflly loaded the policy state dict from {policy_path}')

  def get_chrom_list(self):
    chrom_list = []
    for idx, row in self.peak_qc['chromDB'].chrom[['PeptideModifiedSequence', 'FileName']].drop_duplicates().iterrows():
      pep = row['PeptideModifiedSequence']
      file = row['FileName']
      chrom = self.peak_qc['chromDB'].getChromData(file, pep)
      if chrom is not None:
        chrom_list.append(chrom)
    return chrom_list

  def get_chrom_list_from_sample(self, sample_list, prescreen=0):
    chrom_list = []
    chromDB = self.peak_qc['chromDB']
    for fn, pep in sample_list:
      if prescreen > 0:
        chrom = self.init_chrom(fn, pep, prescreen=prescreen)
      else:
        chrom = chromDB.getChromData(fn, pep)
      if chrom is not None:
        chrom_list.append(chrom)
    return chrom_list

  def init_chrom(self, fn, pep, prescreen=300):
    chromDB = self.peak_qc['chromDB']
    chrom_candidate = chromDB.getChromData(fn, pep)
    if chrom_candidate is None:
      return None
    current_score = self.get_chrom_score(chrom_candidate)['final_reward']
    kde_obj = None
    if self.peak_kde and pep in self.peak_kde and type(self.peak_kde[pep]) != bool:
      kde_obj = self.peak_kde[pep]
    for _ in range(prescreen):
      if kde_obj is not None:
        chrom = chromDB.getChromData(fn, pep, start=kde_obj['start'].resample(1)[0][0], end=kde_obj['end'].resample(1)[0][0])
      else:
        chrom = chromDB.updateChromData(chrom_candidate)
      score = self.get_chrom_score(chrom)['final_reward']
      if score > current_score:
        chrom_candidate = chrom
        current_score = score
    return chrom_candidate

  def get_sample_lists(self, retry_threshold):
    need_repicking = None
    all_sample_list = []
    sample_list = []
    pre_picked = None
    if self.picked_peak_csv:
      need_repicking = dict()
      try:
        pre_picked = read_csv(self.picked_peak_csv)
        if 'FinalReward' in pre_picked:
          failed_records = pre_picked[pre_picked['FinalReward'] < retry_threshold]
          for idx, row in failed_records.iterrows():
            pep = row['Peptide Modified Sequence']
            file = row['File Name']
            if pep not in need_repicking:
                need_repicking[pep] = dict()
            need_repicking[pep][file] = row
      except:
        pre_picked = None
        print('Failed to read peak-picked CSV file')
    for _, row in self.peak_qc['chromDB'].chrom[['PeptideModifiedSequence', 'FileName']].drop_duplicates().iterrows():
      pep = row['PeptideModifiedSequence']
      file = row['FileName']
      all_sample_list.append((file, pep))
      if need_repicking is None or (pep in need_repicking and file in need_repicking[pep]):
        sample_list.append((file, pep))
    return all_sample_list, sample_list, need_repicking, pre_picked
  
  def random_peak(self, chrom_data):
    return self.peak_qc['chromDB'].updateChromData(chrom_data) # random peak start and end when not setting the start and end
  
  def batch_random_peak(self, batch_chrom, inplace=True):
    return [self.random_peak(chrom) for chrom in batch_chrom]
  
  def get_obs_from_chrom(self, chrom_data):
    return _chrom_to_current_obs(chrom_data)
  
  def batch_get_obs_from_chrom(self, batch_chrom):
    return [_chrom_to_current_obs(chrom) for chrom in batch_chrom]
  
  def reset_obs_df(self, obs, chrom):
    return _reset_obs(obs, chrom)
  
  def batch_reset_obs_df(self, batch_obs, batch_chrom):
    return [_reset_obs(obs, chrom) for (obs, chrom) in zip(batch_obs, batch_chrom)]

  def get_chrom_score(self, chrom_data):
    return _calculate_chrom_score(self.peak_qc, 
                                  chrom_data,
                                  pair_ratio_factor=self.pair_ratio_factor,
                                  pbar_heavy_weight=self.pbar_heavy_weight, 
                                  top_n_ion=self.top_n_ion, 
                                  use_weighted_pbar=self.use_weighted_pbar,
                                  intensity_power=self.intensity_power, 
                                  quality_power=self.quality_power,
                                  peak_kde=self.peak_kde)
  
  def step(self, obs_df, chrom_data, current_step, action, history=[], result_queue=None):
    results = _step(obs_df, self.peak_qc, chrom_data, current_step, action, history, self.max_step, 
                    pair_ratio_factor=self.pair_ratio_factor,
                    pbar_heavy_weight=self.pbar_heavy_weight, intensity_power=self.intensity_power, quality_power=self.quality_power,
                    top_n_ion=self.top_n_ion,
                    peak_kde=self.peak_kde, use_weighted_pbar=self.use_weighted_pbar)
    if result_queue:
      result_queue.put(results)
    return results

  def predict(self, obs_mx, returnValue = False):
    device = self.device
    obs = th_tensor(np_array([obs_mx]), dtype=th_float32, device=device)
    with th_no_grad():
      model_policy = self.policy
      features = model_policy.features_extractor(obs)
      latent_pi  = model_policy.mlp_extractor.policy_net(features)
      mean_actions = model_policy.action_net(latent_pi).detach()
      if returnValue:
        latent_vf = model_policy.mlp_extractor.value_net(features)
        value = model_policy.value_net(latent_vf).detach()
    actions = mean_actions.cpu().numpy()
    if returnValue:
      value = value[0][0].cpu().numpy()
      return actions[0], value
    else:
      return actions[0]
  
  def predict_val(self, obs_mx):
    obs = th_tensor(np_array([obs_mx]), dtype=th_float32, device=self.device)
    with th_no_grad():
      model_policy = self.policy
      features = model_policy.features_extractor(obs)
      latent_vf  = model_policy.mlp_extractor.value_net(features)
      value = model_policy.value_net(latent_vf).detach()
    return value[0][0].cpu().numpy()
  
  def run_mcts(self, chrom_init, option, label=None, initial_obs=None, result_queue=None, return_mcts_tree=False):
    cycle = option['cycle'] or 500
    alpha = option['alpha'] or 0.3
    beta = option['beta'] or 0.2
    K = option['K'] or 3**0.5
    selection_noise = option['selection_noise'] or 1.0
    if not label:
      print(f"\r\033[Kprocessing {chrom_init[0]} - {chrom_init[1]}", end='\r')
    else:
      print(f"\r\033[K{label} processing {chrom_init[0]} - {chrom_init[1]}", end='\r', flush=True)
    eval_mode = option['eval_mode'] or 'policy'
    if initial_obs is None:
      initial_obs = self.get_obs_from_chrom(chrom_init)
    mcts_model = MCTS_DPW(alpha=alpha, beta=beta, initial_obs = initial_obs, initial_chrom=chrom_init, env=self, K=K, eval_mode=eval_mode, selection_noise=selection_noise)
    observation = mcts_model.root.state
    chrom = chrom_init
    done = False
    step = 0
    fail_counter = 0
    while not done and (step < self.max_step):
      mcts_model.learn(cycle)
      try:
        action = mcts_model.best_action()
      except Exception as e:
        print(e)
        fail_counter += 1
        if fail_counter > 5:
          break
        else:
          continue
      observation, reward, done, info = self.step(observation, chrom, step, action, history=mcts_model.history)
      chrom = info['chrom']
      step += 1
      if not return_mcts_tree:
        mcts_model.forward(action, observation, chrom, step)
      if done:
        break
    best_peak = mcts_model.get_best_peak_from_history()
    if best_peak:
      best_peak['notes'] = f'{cycle}cycle_' + best_peak['notes']
    if result_queue:
      result_queue.put(best_peak)
    if not return_mcts_tree:
      del mcts_model
    gc_collect()
    if return_mcts_tree:
      return best_peak, mcts_model
    else:
      return best_peak



from os import environ as os_environ
os_environ["OMP_NUM_THREADS"] = "1"
from argparse import ArgumentParser
import pandas as pd
import numpy as np

def check_chrom_tsv(chrom_tsv, output_tsv):
  chrom = pd.read_csv(chrom_tsv, sep='\t')
  chrom.columns = chrom.columns.str.replace(' ', '')
  for mol, mol_data in chrom.groupby('PeptideModifiedSequence'):
    for fn , fn_data in mol_data.groupby('FileName'):
      counter = 0
      ref_times = []
      ref_time_string = ''
      for time_string, data in fn_data.sort_values(['Times', 'IsotopeLabelType', 'PrecursorCharge', 'FragmentIon', 'ProductCharge']).groupby(['Times']):
        counter += 1
        isFragmentValid = True
        if len(ref_times) == 0:
          ref_times = np.array(str(time_string[0]).split(','), dtype=np.float32)
          ref_time_string = time_string[0]
        for _, frag_data in data.groupby(['PrecursorCharge', 'FragmentIon', 'ProductCharge']):
          if len(frag_data) > 2:
            isFragmentValid = False
            break
        if isFragmentValid and counter > 1:
          print(f'Identified {mol} {fn} has multiple chromatograms (count: {counter}) with correct fragments. Trying to use interpolation to align the datasets')
          time_set2 = np.array(str(time_string[0]).split(','), dtype=np.float32)
          chrom.loc[data.index, 'Times'] = ref_time_string
          chrom.loc[data.index, 'Intensities'] = data['Intensities'].str.split(',').apply(lambda x: np.round(np.interp(ref_times, time_set2, np.array(x, dtype=np.float32)), 6)).apply(lambda x: ','.join(map(lambda y: str(y), x)))
        elif not isFragmentValid:
          print(f'Identified {mol} {fn} has multiple chromatograms sharing with the same RT values (count: {counter})')
          added_counter = 0
          for _, frag_data in data.groupby(['PrecursorCharge', 'FragmentIon', 'ProductCharge', 'IsotopeLabelType']):
            pair_counter = counter
            added_counter = len(frag_data)
            for idx, row in frag_data.iterrows():
              chrom.loc[idx, 'FileName'] = chrom.loc[idx, 'FileName'] + f"::{pair_counter}"
              print(row['FileName'] + f'::{pair_counter}', f"{row['PrecursorCharge']}.{row['FragmentIon']}.{row['ProductCharge']} {row['IsotopeLabelType']}")
              pair_counter += 1
          counter = counter + added_counter -1
          
  chrom.to_csv(output_tsv, sep="\t", index=False)
  print(f'Output checked chromatogram tsv file to {output_tsv}')

def main():
  parser = ArgumentParser()
  parser.add_argument("chromatogram_tsv", type=str, help="The chromatogram TSV file path")
  parser.add_argument("output_chrom_tsv", type=str, help="Output chromatogram tsv file with unique filenames for each target sample")
  args = parser.parse_args()
  print("Input Parameters:")
  for arg in vars(args):
    print(arg, '=',getattr(args, arg))
  print('#########################################################')
  check_chrom_tsv(args.chromatogram_tsv, args.output_chrom_tsv)
if __name__ == '__main__':
  main()


from argparse import ArgumentParser
import configparser
import json
from time import time
from os import environ as os_environ
from numpy import split as np_split, arange as np_arange, array as np_array, median as np_median, asarray as np_asarray, inf as np_inf
from pandas import DataFrame
from torch import multiprocessing as th_mp
from torch.cuda import is_available as th_cuda_is_available
from torch.backends import mps as torch_mps
from os.path import join as os_path_join, dirname as os_path_dirname
from MsTargetPeaker import MsTargetPeakerEnv
import MsTargetPeaker
os_environ["OMP_NUM_THREADS"] = "1"

def main():
  th_mp.freeze_support()
  parser = ArgumentParser()
  parser.add_argument('--version', '-v', help='show the version of the tmasque package', action='version', version=MsTargetPeaker.__version__)
  parser.add_argument("chromatogram_tsv", type=str, help="The chromatogram tsv file path")
  parser.add_argument("output_peak_boundary_csv", type=str, help="The output peak_boundary csv file path")
  parser.add_argument("--speed", '-s', type=str, default="SuperFast", help="The speed mode of SuperFast (10X), SuperFast (10X), Faster (5X), Fast (2X), or Standard (1X speed). (default: SuperFast).")
  parser.add_argument("--mode", '-m', type=str, default="MRM", help="The search mode defined in the config file. Default search modes can be MRM or PRM. (default: MRM)")
  parser.add_argument("--config", '-c', type=str, default=None, help="The file path for customized config file. If unset, use the default config. (default: None)")
  parser.add_argument('--picked', type=str, default=None, help="The previously picked boundaries to continue picking. (default: None)")
  parser.add_argument('--process_num', '-p', type=int, default=4, help="The parallel processing number to calculate quality feature values for all peak groups (default: 4)")
  parser.add_argument('--prescreen', '-pre', type=int, default=50, help="Prescreen peaks for better peak boundaries as initial state. (default: 50)")
  parser.add_argument('--internal_standard_type', '-r', type=str, default=None, choices=['heavy', 'light'], help="Set the internal standard reference to heavy or light ions. (default: heavy)") 
  parser.add_argument('--start_round', '-sr', type=int, default=1, help="Specify the start MCTS round set in the config file. This can be helpful when using continuous peak search. (default: 1)") 
  parser.add_argument('--end_round', '-er', type=int, default=7, help="Specify the end MCTS round set in the config file. This can be helpful when using continuous peak search. (default: 7)") 
  parser.add_argument('--device', '-d', type=str, default='cpu', help="Use cpu or cuda device for model peak picking. (default: cpu)")
  args = parser.parse_args()
  print('Input Parameters')
  for arg in vars(args):
    print(arg, getattr(args, arg))
  print('#########################################################')
  start = time()
  config = configparser.ConfigParser()
  if args.config:
    config.read(args.config)
  else:
    config.read(os_path_join(os_path_dirname(__file__), 'MsTargetPeaker.cfg'))
  default_conf = config['DEFAULT']
  speed = default_conf['Speed']
  if 'speed' in args or args.speed:
    speed = args.speed
  search_mode = default_conf['SearchMode']
  if 'search' in args or args.mode:
    search_mode = args.mode
  device = default_conf['Device']
  policy_path = None if default_conf['PolicyPath'] == 'None' else default_conf['PolicyPath']
  process_num = float(default_conf['ParallelProcessNumber'])
  internal_standard_type = default_conf['InternalStandardType']
  max_step = int(default_conf['MaxTimeStep'])
  mcts_param_conf = config['MCTSParam']
  mcts_param = {
    "alpha":  float(mcts_param_conf['Alpha']),
    "beta":   float(mcts_param_conf['Beta']),
    "K":    eval(mcts_param_conf['K']),
    "eval_mode": mcts_param_conf['EvalMode']
  }
  MCTS_cycles = json.loads(config[f'Speed.{speed}']['MCTSCycle'])
  
  searchParams = config[f'SearchParam.{search_mode}']
  use_weighted_PBAR = int(searchParams['UseWeightedPBAR'])
  threshold_list = json.loads(searchParams['Threshold'])
  selection_noise_list = json.loads(searchParams['SelectionNoise'])
  pbar_heavy_weight = json.loads(searchParams['PBARHeavyWeight'])
  pair_ratio_factor_list = json.loads(searchParams['PairRatioFactor'])
  top_n_ion_list = json.loads(searchParams['TopNIon'])
  intensity_power_list = json.loads(searchParams['IntensityPower'])
  quality_power_list = json.loads(searchParams['QualityPower'])
  use_consensus_list = json.loads(searchParams['UseConsensus'])
  consensus_threshold_list = json.loads(searchParams['ConsensusThresholds'])
  use_ref_if_no_consensus_list = json.loads(searchParams['UseRefIfNoConsensus'])
  overwrite_list = json.loads(searchParams['Overwrite'])
  
  output_peak_boundary_csv = args.output_peak_boundary_csv
  th_mp.set_start_method('spawn')
  start_idx = args.start_round - 1
  end_idx = args.end_round - 1
  if 'device' in args or args.device:
    device = args.device
  if device == 'auto':
    device = 'cuda' if th_cuda_is_available() else 'cpu'

  if 'process_num' in args or args.process_num:
    process_num = args.process_num
  if 'internal_standard_type' in args or args.internal_standard_type:
    internal_standard_type = args.internal_standard_type

  current_csv_file = args.picked
  canStop = False
  total_steps = len(MCTS_cycles)
  if total_steps > len(threshold_list):
    total_steps = len(threshold_list)
  for step in range(total_steps):
    if step < start_idx:
      continue
    if step > end_idx:
      break
    print(f"[Round {step+1}] Start.")
    env = MsTargetPeakerEnv(args.chromatogram_tsv, policy_path=policy_path, picked_peak_csv=current_csv_file,
                            max_step=max_step, device=device, 
                            internal_standard_type=internal_standard_type,
                            pair_ratio_factor= pair_ratio_factor_list[step], 
                            pbar_heavy_weight = pbar_heavy_weight[step],
                            intensity_power=intensity_power_list[step],
                            quality_power=quality_power_list[step],
                            top_n_ion=top_n_ion_list[step],
                            use_weighted_pbar = True if use_weighted_PBAR == 1 else False,
                            use_kde=current_csv_file if use_consensus_list[step] == 1 else None,
                            threshold_for_kde=consensus_threshold_list[step],
                            use_ref_if_no_kde= True if use_ref_if_no_consensus_list[step] == 1 else False
                          )
    all_samples, samples_to_process, need_repicking, pre_picked = env.get_sample_lists(threshold_list[step])
    sample_batches = np_split(samples_to_process, np_arange(process_num, len(samples_to_process), process_num))
    prescreen = args.prescreen
    sample_batches = [(batch, prescreen) for batch in sample_batches]
    print(f'Reading peak groups with {prescreen} pre-screens ...')
    with th_mp.Pool(args.process_num) as pool:
      chrom_list = pool.starmap(env.get_chrom_list_from_sample, sample_batches)
      pool.close()
      pool.join()
    chrom_list = [ y for x in list(filter(lambda x: x is not None, chrom_list)) for y in x]
    chrom_list = np_asarray(chrom_list, dtype='object')
    print(f'[Round {step+1}] Number of peak groups (reward < {threshold_list[step]})  to be processed: {len(chrom_list)}/{len(all_samples)} ({round(100*len(chrom_list)/len(all_samples), 2) if len(all_samples) > 0 else 0.00}%)')
    if overwrite_list[step]:
      print(f'[Round {step+1}] Ignore pre-selected peaks.')
    print(f"RUN MCTS-DPW with {mcts_param['eval_mode']} evaluation")
    option = {'cycle': MCTS_cycles[step], 'alpha': mcts_param['alpha'], 'beta': mcts_param['beta'],
                'K': mcts_param['K'], 'eval_mode': mcts_param['eval_mode'], 'selection_noise': selection_noise_list[step]}
    print(f'[Round {step+1} Option] {option}')
    chrom_len = len(chrom_list)
    arg_list = [(chrom, option, f"[Round {step+1} ({idx+1}/{chrom_len})]") for idx, chrom in enumerate(chrom_list)]
    results = []
    total_size = len(arg_list)
    poolsize = total_size if process_num > total_size else process_num
    print(f'Pool size: {poolsize}')
    if poolsize > 0:
      pool = th_mp.Pool(poolsize)
      results = pool.starmap(env.run_mcts, arg_list)
      pool.close()
      pool.join()
      print('Finished MCTS')
    else:
      canStop = True
    #results = list(chain.from_iterable(results))
    result_df = DataFrame(results)
    if not result_df.empty:
      result_df.columns = ['File Name', 'Peptide Modified Sequence', 'Min Start Time', 'Max End Time', 'Type1Reward', 'Type2Reward',
                           'FinalReward', 'PBAR', 'PBARFactor','PairRatioConsistencyMedian', 
                           'PairRatioConsistencyFactor', 'PeakModality', 'PeakModalityFactor', 'IntensityQuantile', 
                           'IntensityQuantileFactor', 'PeakStartFactor', 'PeakEndFactor', 'PeakBoundaryFactor', 'Note']
      non_qualified = result_df[result_df['FinalReward'] < threshold_list[step]]
    else:
      non_qualified = []
    print(f"[Round {step+1}] #Peak Groups with Final Reward < {threshold_list[step]}: {len(non_qualified)}")
    print(f'[Round {step+1}] Outputing picked peaks to {output_peak_boundary_csv} ...')
    if need_repicking is None:
      output_df = result_df
    else:
      output_arr = []
      improved_counts = 0
      improved_scores = []
      for file, pep in all_samples:
        if pep in need_repicking and file in need_repicking[pep]:
          previous_result = need_repicking[pep][file]
          if result_df.empty:
            output_arr.append(previous_result)
          else:
            repick_result = result_df[(result_df['Peptide Modified Sequence'] == pep) & (result_df['File Name'] == file)]
            if len(repick_result) >= 1:
              repicked = repick_result.iloc[0]
              previous_reward = previous_result['FinalReward']
              repicked_reward = repicked['FinalReward']
              if repicked_reward > previous_reward:
                improved_counts += 1
                improved_scores.append(repicked['FinalReward'] - previous_result['FinalReward'])
              if overwrite_list[step] or (repicked_reward >= previous_reward):
                output_arr.append(repicked)
              else:
                output_arr.append(previous_result)
            else:
              output_arr.append(previous_result) #Should not exist
        else:
          previous_result = pre_picked[(pre_picked['Peptide Modified Sequence'] == pep) & (pre_picked['File Name'] == file)] if pre_picked is not None else []
          if len(previous_result) >= 1:
            output_arr.append(previous_result.iloc[0])
      output_df = DataFrame(output_arr)
      improved_scores = np_array(improved_scores)
      if len(improved_scores) > 0:
        improved_mean = improved_scores.mean()
        improved_median = np_median(improved_scores)
        improved_min = improved_scores.min()
        improved_max = improved_scores.max()
        print(f'Improved chromatograms: {improved_counts}')
        print(f'Improved scores => mean: {improved_mean}; median: {improved_median}; min: {improved_min}; max: {improved_max}')
      else:
        print('No improvement')
    print('#########################################################')
    print(f'[Round {step+1}] Score distributions:')
    total_len = len(output_df)
    if total_len > 0:
      ranges = [-1, 0, 0.01, 1, 2, 4, 6, 8, 10, 12]
      for idx, range_start in enumerate(ranges):
        if idx + 1 < len(ranges):
          range_end = ranges[idx + 1]
        else:
          range_end = np_inf
        in_range_data = output_df[(output_df['FinalReward'] >= range_start) & (output_df['FinalReward'] < range_end)]
        in_range_len = len(in_range_data)
        print(f'[{range_start}, {range_end}): {in_range_len} ({ round(100*in_range_len/total_len, 2) }%)')
    print('#########################################################')
    print(f'[Round {step+1}] Output peak boundary csv file to ' + output_peak_boundary_csv)
    current_csv_file = output_peak_boundary_csv
    output_df.to_csv(output_peak_boundary_csv, index=False)
    #output_df.to_csv(f'{output_peak_boundary_csv}.step{step+1}.txt', index=False)
    if canStop:
      break
  end = time()
  print('Total execution time: %.2f seconds' % (end - start))

if __name__ == '__main__':
  main()